from django.contrib.auth import authenticate, login, logout
from django.contrib.auth.decorators import login_required, user_passes_test
from django.contrib.auth.forms import AuthenticationForm
from django.http.response import HttpResponseBadRequest, HttpResponseForbidden
from django.shortcuts import get_object_or_404, redirect, render
from django.http import HttpResponse
from django.db.models import F

from pasapp.forms import ApplicationForm, NewProjectForm, NewUserForm
from pasapp.models import Application, Project, ProjectTag, Tag
from pasapp.utils import contextWithHeader, is_professor, is_student


def index(request):
    '''
    Lists all active projects. Also handles filtering those projects via POST requests.
    URL: /
    '''
    tag_ids, professor_ids = [], []

    if request.method == "POST":
        post_request = request.POST
        # Sort the IDs sent by the filter post request into professor IDs and tag IDs
        for key in post_request.keys():
            if key.startswith("prof"):
                professor_ids.append(int(key.replace("prof", "", 1)))
            elif key.startswith("tag"):
                tag_ids.append(int(key.replace("tag", "", 1)))

        # Within each filter section (e.g. tag filters), we OR each selected option.
        # If no filters are selected for a section, we just return all projects.
        if tag_ids:
            project_tags = ProjectTag.objects.filter(tag_id__in=tag_ids).all()
            tag_projects = {relation.project for relation in project_tags}
        else:
            tag_projects = Project.objects.all()

        if professor_ids:
            professor_projects = Project.objects.filter(
                professor_id__in=professor_ids).all()
        else:
            professor_projects = Project.objects.all()

        # The intersection of the results from each filtering option works like an AND
        projects = tag_projects.intersection(professor_projects)
    else:
        projects = Project.objects.order_by('-title')

    tag_filters = ProjectTag.objects.all()
    # Only show professors who have published a project
    professor_filters = {
        project.professor for project in Project.objects.all()}

    context = {'projects': projects, 'tags': tag_filters, 'professors': professor_filters,
               'selectedTags': tag_ids, 'selectedProfessors': professor_ids}
    return render(request, 'pasapp/projects.html', contextWithHeader(context, request))


def project_details(request, project_id):
    '''
    Shows more details about the given project, including number of applicants
    URL: /project/<id>
    '''
    project = get_object_or_404(Project, pk=project_id)
    num_applicants = Application.objects.filter(project=project_id).count()
    tags = [entry.tag.name for entry in ProjectTag.objects.filter(
        project=project).all()]
    context = {'project': project,
               'numApplicants': num_applicants, 'tags': tags}
    return render(request, 'pasapp/project_details.html', contextWithHeader(context, request))


@login_required
@user_passes_test(is_student)
def apply(request, project_id):
    '''
    Form page that allows a student to apply for a project
    URL: /project/<id>/apply
    '''
    project = get_object_or_404(Project, pk=project_id)
    if request.method == 'POST':
        form = ApplicationForm(request.POST)
        if form.is_valid():
            student = request.user
            message = form.cleaned_data.get('message')
            application = Application(
                project=project, student=student, message=message)
            application.save()
            return redirect(f"/project/{project_id}/")
    form = ApplicationForm()
    context = {'project': project, 'form': form}
    return render(request, 'pasapp/create_application.html', contextWithHeader(context, request))


@login_required
def project_applications(request, project_id):
    '''
    Shows all applications to a project (if you are the professor who owns that project)
    URL: /project/<id>/applications
    '''
    project = get_object_or_404(Project, pk=project_id)
    if not request.user.id == project.professor.id:
        return HttpResponseForbidden()
    applications = Application.objects.filter(
        project=project_id).order_by('-date_created')
    context = {'project': project, 'applications': applications}
    return render(request, 'pasapp/project_applications.html', contextWithHeader(context, request))


@login_required
@user_passes_test(is_professor)
def create_project(request):
    '''
    Form page allowing professors to create a new project
    URL: /project/create
    '''
    if request.method == 'POST':
        form = NewProjectForm(request.POST)
        if form.is_valid():
            title = form.cleaned_data.get('title')
            description = form.cleaned_data.get('description')
            professor = request.user
            status = 'Hidden'
            project = Project(title=title, description=description,
                              status=status, professor=professor)
            project.save()

            tags = form.cleaned_data.get('tags')
            tag_list = tags.split()
            for tag_id_str in tag_list:
                tag_id = int(tag_id_str)
                tag = Tag.objects.get(pk=tag_id)
                if tag is None:
                    raise ValueError("Unknown tag found")
                projectTag = ProjectTag(project=project, tag=tag)
                projectTag.save()

            return redirect("index")
    form = NewProjectForm()
    tags = Tag.objects.all()
    context = {'form': form, 'tags': tags}
    return render(request, 'pasapp/create_project.html', contextWithHeader(context, request))


@login_required
@user_passes_test(is_professor)
def edit_project(request, project_id):
    '''
    Form page that allows a professor to edit and existing project.
    URL: /project/<id>/edit
    '''
    project = get_object_or_404(Project, pk=project_id)
    if not request.user.id == project.professor.id:
        return HttpResponseForbidden()

    if request.method == 'POST':
        form = NewProjectForm(request.POST)
        if form.is_valid():
            title = form.cleaned_data.get('title')
            description = form.cleaned_data.get('description')
            status = form.cleaned_data.get('status')
            project.title = title
            project.description = description
            project.status = status
            project.save()
            return redirect(f'/project/{project_id}/')

    else:
        form = NewProjectForm(initial={
                              'title': project.title, 'description': project.description, 'status': project.status})
    context = {'form': form, 'project': project}
    return render(request, 'pasapp/edit_project.html', contextWithHeader(context, request))


@login_required
def my_applications(request):
    '''
    Displays all of the user's applications, rendered differently for students and professors (see below)
    URL: /applications
    '''
    if is_student(request.user):
        return student_applications_view(request)
    elif is_professor(request.user):
        return professor_applications_view(request)
    else:
        return HttpResponse("Only students and professors may view this page.")


def student_applications_view(request):
    '''
    Displays all of the student's project applications and allows them to set rankings (see update_priorities below)
    '''
    applications = Application.objects.filter(student=request.user).order_by(
        F('priority').asc(nulls_last=True), '-date_created')
    context = {'applications': applications}
    return render(request, 'pasapp/student_applications.html', contextWithHeader(context, request))


def professor_applications_view(request):
    '''
    Displays all applications to the professor's projects, grouped by project
    '''
    projects = Project.objects.filter(professor=request.user)
    sections = {project: Application.objects.filter(
        project=project).order_by('-date_created') for project in projects}
    context = {'sections': sections}
    return render(request, 'pasapp/professor_applications.html', contextWithHeader(context, request))


@login_required
@user_passes_test(is_student)
def update_priorities(request):
    '''
    Endpoint for updating the priority values of a student's applications. POST only.
    URL: /applications/update_priorities
    '''
    if request.method == 'POST':
        post_request = request.POST
        for field in post_request:
            if field.startswith("priority"):
                application_id = post_request[field]
                priority = int(field.replace("priority", "", 1))

                duplicate_priority_applications = Application.objects.filter(
                    priority=priority)
                for application in duplicate_priority_applications:
                    application.priority = None
                    application.save()

                if application_id == '':
                    continue
                application = Application.objects.get(pk=application_id)
                if application is not None:
                    application.priority = priority
                    application.save()
        return redirect("/applications/")

    else:
        return HttpResponseBadRequest("This endpoint only supports POST requests.")


def register(request):
    '''
    Form page for registering a new user.
    URL: /register
    '''
    if request.method == 'POST':
        form = NewUserForm(request.POST)
        if form.is_valid():
            user = form.save()
            login(request, user)
            return redirect("index")
    else:
        form = NewUserForm()
    context = {'form': form}
    return render(request, 'pasapp/register.html', contextWithHeader(context, request))


def login_view(request):
    '''
    Form page for logging a user in. Mostly used as a redirect if the user attempts to access
    a page that requires login.
    URL: /login
    '''
    if request.method == 'POST':
        form = AuthenticationForm(data=request.POST)
        print(form)
        if form.is_valid():
            username = form.cleaned_data.get('username')
            password = form.cleaned_data.get('password')
            print(password)
            user = authenticate(username=username, password=password)
            if user is not None:
                login(request, user)
                return redirect('index')
        print(form.non_field_errors())
    else:
        form = AuthenticationForm()
    return render(request, 'pasapp/login.html', {'form': form})


def logout_view(request):
    '''
    Endpoint for logout requests. Logs the current user out and returns to the index page.
    URL: /logout
    '''
    logout(request)
    return redirect('index')
